function extractColor(inputFileName, smoothingSigma, maxNFrames, saveVisualizations, forceGrayscale)
%EXTRACTCOLOR Extract image features from video.
%   EXTRACT*(INPUTFILENAME, SMOOTHINGSIGMA, MAXNFRAMES, SAVEVISUALIZATIONS, FORCEGRAYSCALE)
%     INPUTFILENAME      Input video file
%     MAXNFRAMES         Optional, maximum number of frames of the input video to process (useful to run faster tests only on the beginning of a sequence)
%     SMOOTHINGSIGMA     Sigma of Gaussian (blurring) filter to apply; leave empty to skip this step.
%     SAVEVISUALIZATIONS Save videos that show the extracted features.
%     FORCEGRAYSCALE

% Author: Damien Teney

featureType = 0;
[filePath fileName fileExtension] = fileparts(inputFileName);
if isempty(filePath), filePath = '.'; end

% Delete existing results (from previous runs of the function)
delete([filePath filesep fileName '_features' int2str(featureType) '*']);

%--------------------------------------------------------------------------
% Load video frames
%--------------------------------------------------------------------------
printf('Loading video...'); tic();
% Get total number of frames of the video
[trash width height nFrames] = loadVideo(inputFileName, []);
nFrames = min(nFrames, maxNFrames); % Limit the number of frames
[video width height ~] = loadVideo(inputFileName, 1 : nFrames);
printf(' %.2fs\n', toc());

% Smooth the video (2D filter)
if smoothingSigma > 0
  printf('Smoothing the video...'); tic();
  filter = fspecial('gaussian', 2 * ceil(3 * smoothingSigma) + 1, smoothingSigma);
  for frameId = 1:nFrames
    video(:, :, frameId) = imfilter(video(:, :, frameId), filter, 'symmetric');
  end
  printf(' %.2fs\n', toc());

  % Save visualizations of the smoothed video
  if saveVisualizations
    printf('Saving a visualization...'); tic();
    outputFileName = fullfile(filePath, [fileName '_features' int2str(featureType) '.avi']);
    %saveVideo(video, outputFileName, [0 1]);
    saveVideo(video, outputFileName, []);
    printf(' %.2fs\n', toc());
  end
end

%--------------------------------------------------------------------------
% Transform color values into histograms
%--------------------------------------------------------------------------
features = video; % Rename but also keep a copy of the video RGB values (as will be saved)

% Convert RGB values to Lab values
for frameId = 1:nFrames
  video(:, :, frameId, :) = applycform(squeeze(video(:, :, frameId, :)), makecform('srgb2lab'));
end

histBins = reshape(video, [], 3);
clear video;
histBins(:, 2) = 50 + histBins(:, 2) ./ 2; % Get A values in [0, 100]
histBins(:, 3) = 50 + histBins(:, 3) ./ 2; % Get B values in [0, 100]
histBins = histBins ./ 100; % Get all values in [0, 1]

% Make 3 separate 1D histograms
N_BINS = 10; % Number of bins for each of the 3 components
histBins = ceil(histBins * N_BINS); % Get indices of histogram bins
histBins = max(min(histBins, N_BINS - 1), 0); % Cap values in [0, N_BINS[ (necessary in case of numerical errors)

nPixels = height * width * nFrames;
histograms = zeros(nPixels, N_BINS * 3, 'double');

% Increment the appropriate bins of the histogram
binsToIncrement = sub2ind(size(histograms), 1:nPixels, 1 + 0 * N_BINS + histBins(:, 1)'); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), 1:nPixels, 1 + 1 * N_BINS + histBins(:, 2)'); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), 1:nPixels, 1 + 2 * N_BINS + histBins(:, 3)'); histograms(binsToIncrement) = 1;

% Increment neighbouring bins
%%{
%for offset = [-1 +1]
for offset = [-1]
  histBins2 = histBins + offset;
  histBins2(:) = max(histBins2(:), 0); % Remove values < 0
  histBins2(:) = min(histBins2(:), N_BINS-1); % Remove values >= N_BINS
  binsToIncrement = sub2ind(size(histograms), 1:nPixels, 1 + 0 * N_BINS + histBins2(:, 1)'); histograms(binsToIncrement) = 1;
  binsToIncrement = sub2ind(size(histograms), 1:nPixels, 1 + 1 * N_BINS + histBins2(:, 2)'); histograms(binsToIncrement) = 1;
  binsToIncrement = sub2ind(size(histograms), 1:nPixels, 1 + 2 * N_BINS + histBins2(:, 3)'); histograms(binsToIncrement) = 1;
end
%}

if nargin >= 5 && forceGrayscale
  histograms = histograms(:, 1:N_BINS); % Keep only the histogram bins of the luminance
else
  histograms = histograms ./ 3; % Ensure each row sums to 1 (we treat the concatenation of 3 histograms as 1 single normalized histogram)
end

histograms = bsxfun(@times, histograms, 1 ./ sum(histograms, 2)); % Normalize so that all rows sum to 1

%--------------------------------------------------------------------------
% Save the results
%--------------------------------------------------------------------------
assert(~any(isinf(histograms(:))));
assert(~any(isnan(histograms(:))));

printf('Saving the results...'); tic();
outputFileName = fullfile(filePath, [fileName '_features' int2str(featureType) '.mat']);
save(outputFileName, 'features', 'histograms', 'height', 'width', 'nFrames', '-v7.3');
printf(' %.2fs\n', toc());
