function extractHoMEs(inputFileName, smoothingSigma, scales, speeds, maxNFrames, discardBorders, filterType, sumAlongPlanes)
%EXTRACT* Extract image features from video.
%   EXTRACTHOMES(INPUTFILENAME, SMOOTHINGSIGMA, SCALES, SPEEDS, MAXNFRAMES, SAVEVISUALIZATIONS, DISCARDBORDERS
%     INPUTFILENAME      Input video file
%     MAXNFRAMES         Optional, maximum number of frames of the input video to process (useful to run faster tests only on the beginning of a sequence)
%     SMOOTHINGSIGMA     Sigma of Gaussian (blurring) filter to apply; leave empty to skip this step.
%     SCALES             Vector of integer values, e.g. 2 for half-size. For example: [1] for only 1 scale.
%     SPEEDS             Vector of values in pixels per frame.
%                        For example:
%                        speeds = [0 0.5 1.0 2.0 3.0 +inf]; % A few speeds, including static (0) and flicker (+inf)
%                        speeds = [0 : 0.25 : 1.5];
%                        speeds = [0 : 0.1 : 1.0 +inf]; % Slow speeds and flicker
%                        speeds = [0 : 0.1 : 1.0]; % Slow speeds
%                        speeds = [0 : 0.2 : 2.0]; % Medium speeds
%                        speeds = [0 : 0.5 : 5.0]; % Medium speeds
%                        speeds = [0 : 1.0 : 10.0]; % Fast speeds
%                        speeds = 0.5 .* (logspace(0, 1, 10) - 1); % Speeds 0-4.5, log spaced
%       speeds = logspace(0, 1, 15) - 1; % Speeds 0-9, log spaced
%     DISCARDBORDERS     Do not save features in the (unreliable) borders of the video volume.
%
%    Reference:
%      Segmentation of Dynamic Scenes with Distributions of Spatiotemporally Oriented Energies
%      Damien Teney, Matthew Brown, BMVC, 2014.

% Author: Damien Teney

if ~sumAlongPlanes
  featureType = 20;
else
  featureType = 21;
end
[filePath fileName fileExtension] = fileparts(inputFileName);
if isempty(filePath), filePath = '.'; end

%--------------------------------------------------------------------------
% Load video frames
%--------------------------------------------------------------------------
printf('Loading video...'); tic();
% Get total number of frames of the video
[trash width height nFrames] = loadVideo(inputFileName, []);
nFrames = min(nFrames, maxNFrames); % Limit the number of frames
[video width height ~] = loadVideo(inputFileName, 1 : nFrames);
video = convertVideoToGrayscale(video);
printf(' %.2fs\n', toc());

% Smooth the video (2D filter)
if smoothingSigma > 0
  printf('Smoothing the video...'); tic();
  filter = fspecial('gaussian', 2 * ceil(3 * smoothingSigma) + 1, smoothingSigma);
  for frameId = 1:nFrames
    video(:, :, frameId) = imfilter(video(:, :, frameId), filter, 'symmetric');
  end
  printf(' %.2fs\n', toc());
end

%--------------------------------------------------------------------------
% Main code
%--------------------------------------------------------------------------
% Missing parameters: set default values
if isempty(scales) 
  scales = [1];
end
if isempty(speeds)
  speeds = [0 : 0.2 : 2.0];
end

nOrientations = 16;
orientationAngles = [0 : nOrientations-1] * 2 * pi / nOrientations;
orientations = [ cos(orientationAngles)' sin(orientationAngles)' ];

filterId = 1;
filterResponses = zeros(height, width, nFrames, 0);
filterInfos = zeros(0, 3); % Store which filter corresponds to which scale/speed/orientation

for scaleId = 1:length(scales)
  video2 = resizeVideo(video, 1 / scales(scaleId));
  tmp = zeros(size(video2)); % Initialize variable to store the response of the filters

  % Precompute basis images
  switch filterType
    case 'G2H2'
      G2Basis = init3DG2(video2);
      H2Basis = init3DH2(video2);
      N = 3; % Number of steering orientations (in the frequency domain plane) that we sum together (order of Gaussian + 1)
    case 'G3H3'
      G3Basis = init3DG3(video2);
      H3Basis = init3DH3(video2);
      N = 4;
    case 'G2'
      G2Basis = init3DG2(video2);
      N = 3;
    case 'G3'
      G3Basis = init3DG3(video2);
      N = 4;
    case 'G4'
      G4Basis = init3DG4(video2);
      N = 5;
    otherwise
      error('Unknown filter type !');
  end
  if ~sumAlongPlanes
    % Do not sum filter responses that lie in planes in the frequency domain, but rather keep individual filter responses
    N = 1;
  end

  for speedId = 1:length(speeds)
    for orientationId = 1:size(orientations, 1)
      printf('\tScale %.1f, speed %.1f, orientation [%.1f %.1f]\n', scales(scaleId), speeds(speedId), orientations(orientationId, 1), orientations(orientationId, 2)); tic();

      % Get the steering orientations of the filters to apply (and to sum the responses of)
      %--------------------------------------------------------------------
      % Static
      %--------------------------------------------------------------------
      if speeds(speedId) == 0
        if N == 1
          steeringOrientations = [ orientations(orientationId, 1) orientations(orientationId, 2) 0 ];
        else
          if orientationId > 1
            % Orientation is not relevant for the static response; only do it for one
            %filterResponses(:, :, :, filterId) = 0;
            filterResponses(:, :, :, filterId) = filterResponses(:, :, :, filterId - 1);
            filterInfos(filterId, :) = [scaleId speedId orientationId];
            filterId = filterId + 1;
            continue;
          end
          n = [0 0 1];
          a = cross(n, [1 0 0]); a = a ./ norm(a);
          b = cross(n, a); b = b ./ norm(b);
          steeringOrientations = [ cos([1:N]' .* pi / N) * a + sin([1:N]' .* pi / N) * b ];
        end

      %--------------------------------------------------------------------
      % Flicker
      %--------------------------------------------------------------------
      elseif speeds(speedId) == +inf
        if N == 1
          steeringOrientations = [0 0 1];
        else
          if orientationId > size(orientations, 1) / 2
            % Only half of the orientations (on [0,180[ and not the full [0,360[) are relevant for the flicker response
            % Only do it for the first half (assuming they were defined in order !!)
            %filterResponses(:, :, :, filterId) = 0;
            filterResponses(:, :, :, filterId) = filterResponses(:, :, :, filterId - 1);
            filterInfos(filterId, :) = [scaleId speedId orientationId];
            filterId = filterId + 1;
            continue;
          end
          n = [orientations(orientationId, 1) ...
               orientations(orientationId, 2) ...
               0];
          n = n ./ norm(n);
          otherOrientationId = orientationId - 1; if orientationId == 1, otherOrientationId = orientationId + 1; end
          a = cross(n, [orientations(otherOrientationId, 1) orientations(otherOrientationId, 2) 0]); a = a ./ norm(a);
          b = cross(n, a); b = b ./ norm(b);
          steeringOrientations = [ cos(pi .* [0:N-1]'./N) * a + sin(pi * [0:N-1]'./N) * b ];
        end

      %--------------------------------------------------------------------
      % "Normal" motion
      %--------------------------------------------------------------------
      else
        if N == 1
          steeringOrientations = [ orientations(orientationId, 1) orientations(orientationId, 2) -speeds(speedId) ];
        else
          n = [speeds(speedId) * orientations(orientationId, 1) ...
               speeds(speedId) * orientations(orientationId, 2) ...
               1];
          n = n ./ norm(n);
          a = cross(n, [1 0 0]); a = a ./ norm(a);
          b = cross(n, a); b = b ./ norm(b);
          steeringOrientations = [ cos(pi .* [0:N-1]'./N) * a + sin(pi * [0:N-1]'./N) * b ];
        end
      end

      tmp(:) = 0;
      for i = 1:N
        steeringOrientation = steeringOrientations(i, 1:3);
        steeringOrientation = steeringOrientation ./ norm(steeringOrientation); % Normalize to unit norm
        switch filterType
          case 'G2H2'
            videoG = steer3DG2(steeringOrientation, G2Basis);
            videoH = steer3DH2(steeringOrientation, H2Basis);
            tmp = tmp + videoG.^2 + videoH.^2;
          case 'G3H3'
            videoG = steer3DG3(steeringOrientation, G3Basis);
            videoH = steer3DH3(steeringOrientation, H3Basis);
            tmp = tmp + videoG.^2 + videoH.^2;
          case 'G2'
            videoG = steer3DG2(steeringOrientation, G3Basis);
            tmp = tmp + videoG.^2;
          case 'G3'
            videoG = steer3DG3(steeringOrientation, G3Basis);
            tmp = tmp + videoG.^2;
          case 'G4'
            videoG = steer3DG4(steeringOrientation, G4Basis);
            tmp = tmp + videoG.^2;
        end
      end

      % Put the result in a big matrix with the response of all filters
      filterResponses(:, :, :, filterId) = resizeVideo(tmp, [size(video, 1) size(video, 2)]); % Restore to original video scale
      filterInfos(filterId, :) = [scaleId speedId orientationId];

      filterId = filterId + 1;
    end % For each orientation
  end % For each speed
end % For each scale

clear video video2 tmp;

nFilters = size(filterInfos, 1);

%--------------------------------------------------------------------------
% Determine the borders of the video frames (invalid parts of the filtering)
%--------------------------------------------------------------------------
if discardBorders
  borderPixels = false(height, width, nFrames);
  s = ceil(4 * max(scales)); % Filters half size in spatial dimensions
  %t = ceil(4); % Filters half size in time dimension
  t = ceil(3); % Filters half size in time dimension
  borderPixels(1 : s, :, :) = 1;
  borderPixels(:, 1 : s, :) = 1;
  borderPixels(:, :, 1 : t) = 1;
  borderPixels(end+1-s : end, :, :) = 1;
  borderPixels(:, end+1-s : end, :) = 1;
  borderPixels(:, :, end+1-t : end) = 1;

  % Remove the values at the borders
  filterResponses(1 : s, :, :, :) = 0;
  filterResponses(:, 1 : s, :, :) = 0;
  filterResponses(:, :, 1 : t, :) = 0;
  filterResponses(end+1-s : end, :, :, :) = 0;
  filterResponses(:, end+1-s : end, :, :) = 0;
  filterResponses(:, :, end+1-t : end, :) = 0;
else
  borderPixels = [];
end

%--------------------------------------------------------------------------
% Make histograms of the responses of the filters
%--------------------------------------------------------------------------
histograms = reshape(filterResponses, [], nFilters);
clear filterResponses;

% Rescale histograms so that peaks are more "peaky"
%%{
% Normalize so that the max of each row is 1 (makes subsequent scaling more consistent)
blackPixels = (max(histograms, [], 2) == 0);
histograms = bsxfun(@times, histograms, 1 ./ max(histograms, [], 2));
% Make more peaky
histograms = exp((histograms - 1) .* 1000);
histograms(blackPixels, :) = 1; % Set uniform histograms to black pixels (otherwise would have NaN values)
%}

histograms = bsxfun(@times, histograms, 1 ./ sum(histograms, 2)); % Normalize so that all rows sum to 1

histograms(borderPixels(:), :) = 0; % Empty the histograms of border pixels (the normalization would have generated NaN vaues)

%--------------------------------------------------------------------------
% Save the results
%--------------------------------------------------------------------------
assert(~any(isinf(histograms(:))));
assert(~any(isnan(histograms(:))));

printf('Saving the results...'); tic();
delete([filePath filesep fileName '_features' int2str(featureType) '*']); % Delete existing results (from previous runs of the functions)
outputFileName = fullfile(filePath, [fileName '_features' int2str(featureType) '.mat']);
save(outputFileName, 'histograms', 'borderPixels', 'height', 'width', 'nFrames', 'nOrientations', 'orientationAngles', 'speeds', 'scales', '-v7.3');
printf(' %.2fs\n', toc());
