function extractOpticalFlow(inputFileName, frameSpacing, maxNFrames, saveVisualizations)
%EXTRACTOPTICALFLOW Extract image features from video.
%   EXTRACTOPTICALFLOW(INPUTFILENAME, FRAMESPACING, MAXNFRAMES, SAVEVISUALIZATIONS)
%     INPUTFILENAME      Input video file
%     MAXNFRAMES         Optional, maximum number of frames of the input video to process (useful to run faster tests only on the beginning of a sequence)
%     FRAMESPACING       Spacing (in number of frames) between the pair of frames used to compute the optical flow. Usually equals to 1, so that the flow is compute on successive frames. The flow may be more stable with a larger value, but is only appropriate if the motion is fairly constant over time.
%     SAVEVISUALIZATIONS Save videos that show the extracted features.

% Author: Damien Teney

featureType = 30;
[filePath fileName fileExtension] = fileparts(inputFileName);
if isempty(filePath), filePath = '.'; end

%--------------------------------------------------------------------------
% Load video frames
%--------------------------------------------------------------------------
printf('Loading video...'); tic();
% Get total number of frames of the video
[trash width height nFrames] = loadVideo(inputFileName, []);
nFrames = min(nFrames, maxNFrames); % Limit the number of frames
[video width height ~] = loadVideo(inputFileName, 1 : nFrames);
video = convertVideoToGrayscale(video);
printf(' %.2fs\n', toc());

%--------------------------------------------------------------------------
% Main code
%--------------------------------------------------------------------------
if isempty(frameSpacing) % Parameter not given, set default value
  frameSpacing = 1;
end

% Check that the parameter is valid
assert(frameSpacing >= 1);
if nFrames < (1 + frameSpacing)
  error('Video too short for the chosen frameSpacing !');
end

flowUV = zeros(height, width, nFrames, 2);

printf('Current frame: ');
for frameId = (1 + frameSpacing) : (nFrames - frameSpacing)
  printf('%u ', frameId);
  frame0 = video(:, :, frameId - frameSpacing);
  frame1 = video(:, :, frameId);
  frame2 = video(:, :, frameId + frameSpacing);

  % Compute the deformation
  alpha = 0.012;
  ratio = 0.75;
  minWidth = 20;
  nOuterFPIterations = 7;
  nInnerFPIterations = 1;
  nSORIterations = 30;
  parameters = [alpha ratio minWidth nOuterFPIterations nInnerFPIterations nSORIterations];

  % Forward flow with previous AND next frames
  [u0 v0 trash] = Coarse2FineTwoFrames(im2double(frame0), im2double(frame1), parameters); % For each pixel of frame1, how it moves to get to frame0
  uv0 = cat(3, u0, v0); % height x width x 2 matrix
  [u1 v1 trash] = Coarse2FineTwoFrames(im2double(frame1), im2double(frame2), parameters); % For each pixel of frame1, how it moves to get to frame2
  uv1 = cat(3, u1, v1); % height x width x 2 matrix
  %uv = (uv0 + uv1) ./ 2; % Average
  uv = uv0; % Just keep one of the two

  % Scale back the values to have them in pixel per ONE frame (we may have computed between more distant frames if frameSpacing > 1)
  uv = uv ./ frameSpacing;

  % Debug display
  %{
  figure(); imshow(frame1); figure(); imshow(frame2);
  figure(); imshow(-uv0(:, :, 1), []);
  figure(); imshow( uv1(:, :, 1), []);
  figure(); imshow( uv(:, :, 1), []);
  figure(); imshow( uv2b(:, :, 1), []);
  figure(); imshow(flowDifference, []);
  %}

  flowUV(:, :, frameId, 1:2) = uv;
end
printf('\n');

% Normalize magnitude per frame for overall motion
%{
for frameId = 1:nFrames
  frameUs = flowUV(:, :, frameId, 1);
  frameVs = flowUV(:, :, frameId, 2);
  flowUV(:, :, frameId, 1) = flowUV(:, :, frameId, 1) - mean(frameUs(:));
  flowUV(:, :, frameId, 2) = flowUV(:, :, frameId, 2) - mean(frameVs(:));
  %flowUV(:, :, frameId, 1) = flowUV(:, :, frameId, 1) - median(frameUs(:));
  %flowUV(:, :, frameId, 2) = flowUV(:, :, frameId, 2) - median(frameVs(:));
end
%}

% Turn UV maps to orientation/magnitude maps
flowPolar = flowUvToPolar(flowUV);

% Normalize magnitude per frame for magnitude
%{
for frameId = 1:nFrames
  frameMagnitudes = flowPolar(:, :, frameId, 2);
  flowPolar(:, :, frameId, 2) = flowPolar(:, :, frameId, 2) ./ max(frameMagnitudes(:));
  %flowPolar(:, :, frameId, 2) = flowPolar(:, :, frameId, 2) ./ median(frameMagnitudes(:));
end
%}

% Set the list of parts of the input data for which we haven't extracted flow
borderPixels = false(height, width, nFrames);
t = frameSpacing;
borderPixels(:, :, 1 : t) = 1;
borderPixels(:, :, end+1-t : end) = 1;
% Remove the values at the borders
flowUV(:, :, 1 : t, :) = 0;
flowUV(:, :, end+1-t : end, :) = 0;
flowPolar(:, :, 1 : t, :) = 0;
flowPolar(:, :, end+1-t : end, :) = 0;

%--------------------------------------------------------------------------
% Turn the flow values into histogram bins
%--------------------------------------------------------------------------
% Get indices of histogram bins
% Orientation
N_BINS_ORI = 16;
orientations = flowPolar(:, :, :, 1); % Values in [0, 2*pi[
orientations = orientations(:);
histBin = orientations ./ (2 * pi); % Values in [0, 1[
assert(all(histBin(:) >= 0));
assert(all(histBin(:) < 1));
histBin = ceil(N_BINS_ORI * histBin); % Indices of histogram bins
histBin = max(min(histBin, N_BINS_ORI - 1), 0); % Cap values in [0, N_BINS[ (necessary in case of numerical errors)
histBinOri = histBin;

% Magnitude
N_BINS_MAGNITUDE = 1;
magnitudes = flowPolar(:, :, :, 2); % Reshape as a long vector
magnitudes = magnitudes(:);
maxMagnitude = prctile(magnitudes(magnitudes(:) > 0), 95); % Upper treshold on magnitude; the values above this are considered outliers
%maxMagnitude = 1.0;
magnitudes = min(magnitudes, maxMagnitude); % Cap those outliers to this "max" threshold value
histBin = magnitudes ./ maxMagnitude; % Values in [0, 1[
assert(all(histBin(:) >= 0));
assert(all(histBin(:) <= 1));
histBin = ceil(N_BINS_MAGNITUDE * histBin); % Indices of histogram bins
histBin = max(min(histBin, N_BINS_MAGNITUDE - 1), 0); % Cap values in [0, N_BINS[ (necessary in case of numerical errors)
histBinMagnitude = histBin;

nPixels = height * width * nFrames;
histograms = zeros(nPixels, N_BINS_ORI, N_BINS_MAGNITUDE);

% Increment the appropriate bins of the histogram
% Option 1: one single bin
%%{
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', histBinOri + 1, histBinMagnitude + 1); histograms(binsToIncrement) = 1;
histograms = reshape(histograms, nPixels, []);
%}
keyboard
% Option 2: neighbouring bins included
%{
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri - 1, N_BINS_ORI) + 1, mod(histBinMagnitude - 1, N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri    , N_BINS_ORI) + 1, mod(histBinMagnitude - 1, N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri + 1, N_BINS_ORI) + 1, mod(histBinMagnitude - 1, N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri - 1, N_BINS_ORI) + 1, mod(histBinMagnitude    , N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri    , N_BINS_ORI) + 1, mod(histBinMagnitude    , N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri + 1, N_BINS_ORI) + 1, mod(histBinMagnitude    , N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri - 1, N_BINS_ORI) + 1, mod(histBinMagnitude + 1, N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri    , N_BINS_ORI) + 1, mod(histBinMagnitude + 1, N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
binsToIncrement = sub2ind(size(histograms), (1:nPixels)', mod(histBinOri + 1, N_BINS_ORI) + 1, mod(histBinMagnitude + 1, N_BINS_MAGNITUDE) + 1); histograms(binsToIncrement) = 1;
histograms = reshape(histograms, nPixels, []);
histograms = bsxfun(@times, histograms, 1 ./ sum(histograms, 2)); % Normalize so that all rows sum to 1
%}

histograms(borderPixels(:), :) = 0; % Empty the histograms of border pixels (the normalization would have generated NaN vaues)

%--------------------------------------------------------------------------
% Save visualizations
%--------------------------------------------------------------------------
% Delete existing results (from previous runs of the functions)
delete([filePath filesep fileName '_features' int2str(featureType) '*']);

if saveVisualizations
  printf('Saving visualizations...'); tic();

  % U
  tmp = flowUV(:, :, :, 1);
  tmp = tmp ./ maxMagnitude;
  outputFileName = sprintf('%s%s%s_features%u_u.avi', filePath, filesep, fileName, featureType);
  saveVideo(tmp, outputFileName, [-1 1]);

  % V
  tmp = flowUV(:, :, :, 2);
  tmp = tmp ./ maxMagnitude;
  outputFileName = sprintf('%s%s%s_features%u_v.avi', filePath, filesep, fileName, featureType);
  saveVideo(tmp, outputFileName, [-1 1]);

  % Orientation
  tmp = flowPolar(:, :, :, 1) ./ (2 * pi);
  outputFileName = sprintf('%s%s%s_features%u_orientation.avi', filePath, filesep, fileName, featureType);
  saveVideo(tmp, outputFileName, [0 1]);

  % Magnitude
  tmp = flowPolar(:, :, :, 2) ./ maxMagnitude;
  outputFileName = sprintf('%s%s%s_features%u_magnitude.avi', filePath, filesep, fileName, featureType);
  saveVideo(tmp, outputFileName, [0 1]);

  % "Color-wheel" flow
  tmp = zeros(height, width, nFrames, 3);
  tmp(:, :, :, 1) = flowPolar(:, :, :, 1) ./ (2 * pi); % Hue (orientation)
  tmp(:, :, :, 2) = flowPolar(:, :, :, 2) ./ maxMagnitude; % Saturation (magnitude)
  tmp(:, :, :, 3) = 1; % Value
  tmp = min(tmp, 1); % Cap values to maximum 1
  for frameId = 1:nFrames
    tmp(:, :, frameId, :) = hsv2rgb(squeeze(tmp(:, :, frameId, :)));
  end
  outputFileName = sprintf('%s%s%s_features%u.avi', filePath, filesep, fileName, featureType);
  saveVideo(tmp, outputFileName, [0 1]);

  printf(' %.2fs\n', toc());
end

%--------------------------------------------------------------------------
% Save the results
%--------------------------------------------------------------------------
assert(~any(isinf(histograms(:))));
assert(~any(isnan(histograms(:))));

printf('Saving the results...'); tic();
outputFileName = fullfile(filePath, [fileName '_features' int2str(featureType) '.mat']);
save(outputFileName, 'histograms', 'flowUV', 'borderPixels', 'height', 'width', 'nFrames', '-v7.3');
printf(' %.2fs\n', toc());
