function visualizeFeatures(inputFile, featureType)
%VISUALIZEFEATURES Visualize image features.

% Author: Damien Teney

figureId = -1;
figureIdBars = -1;
figureIdPolarMap = -1;

visualizePeakOrientation = false;
scaledFeatures = false;

%--------------------------------------------------------------------------
% Load the data
%--------------------------------------------------------------------------
[filePath fileName fileExtension] = fileparts(inputFile);
if isempty(filePath), filePath = '.'; end

% Load video
if ~exist(inputFile, 'file')
  error('Cannot find video file !');
end
[video width height nFramesVideo] = loadVideo(inputFile);

% Load image features
if nargin < 2, featureType = 21; end % Argument not given, set default value
inputFileFeatures = fullfile(filePath, [fileName '_features' int2str(featureType) '.mat']);
if ~exist(inputFileFeatures, 'file')
  error('Cannot find file with image features !');
end
load(inputFileFeatures, 'histograms', 'height', 'width', 'nFrames', 'nOrientations', 'speeds');
featureDimension = size(histograms, 2);
nSpeeds = ceil(featureDimension / nOrientations); % Round to lower integer, maybe there is an extra column with no defined orientation
%assert(round(nSpeeds) == nSpeeds);
imageFeatures = histograms; % Rename
clear histograms;
if size(imageFeatures, 2) < (nOrientations * nSpeeds)
  % Add empty columns so that it can be displayed as a 2D histogram
  imageFeatures(:, (nOrientations * nSpeeds)) = 0;
end

nFrames = min(nFramesVideo, nFrames);

frameId = ceil(nFrames / 2); % Start at the middle of the sequence

figureId = figure(); % Create an empty figure
printf('Press a key:\n');
printf('\tleft/right Previous/next frame\n');
printf('\tr Select a region to show the features of\n');
printf('\tv View a "colorwheel" representation of the peak motion energy\n');
printf('\tz Toggle scaled / unscaled (=raw) feature histograms\n');
printf('\ts Save the figures\n');
printf('\tq Quit\n\n');
while true
  set(figureId, 'Name', ['Frame ' int2str(frameId) '/' int2str(nFrames)], 'NumberTitle', 'off');
  currentFrame = squeeze(video(:, :, frameId, :));
  if ~visualizePeakOrientation
    figure(figureId);
    imshow(currentFrame);
  else
    imageFeaturesFrame = reshape(imageFeatures, height, width, nFrames, []);
    imageFeaturesFrame = squeeze(imageFeaturesFrame(:, :, frameId, :));
    imageFeaturesFrame = reshape(imageFeaturesFrame, (height * width), []);

    % Get the max ori/speed
    %imageFeaturesFrame = imageFeaturesFrame(:, 1 : (end - 1)); % Remove the "flicker" bin
    [trash maxBinIds] = max(imageFeaturesFrame, [], 2);
    [maxOris maxSpeeds] = ind2sub([nOrientations nSpeeds-1], maxBinIds);
    % Get values in [0, 1]
    maxOris = (maxOris - 1) / nOrientations;
    maxSpeeds = (maxSpeeds - 1) / (nSpeeds-1);

    visualization = zeros(height, width, 3);
    visualization(:, :, 1) = reshape(maxOris, height, width); % Hue (orientation)
    visualization(:, :, 2) = reshape(maxSpeeds, height, width); % Saturation (speed)
    visualization(:, :, 3) = 1; % Value
    visualization = min(visualization, 1); % Cap values to maximum 1
    visualization = hsv2rgb(visualization);

    figure(figureId);
    imshow(visualization, [0 1]);
  end

  %------------------------------------------------------------------------
  % Display histograms of a segment
  %------------------------------------------------------------------------
  if ishandle(figureIdBars) % The user has picked a pixel and histograms are displayed
    % Mark the pixel/region picked by the user
    figure(figureId); hold on;
    if y+rectHeight > height, rectHeight = height-y; end % Fix out of bound indices
    if x+rectWidth  > width , rectWidth  = width -x; end
    if rectHeight == 1 && rectWidth == 1
      plot(x, y, 'rs', 'MarkerSize', 8);
    else
      plot([x x+rectWidth x+rectWidth x x], [y y y+rectHeight y+rectHeight y], 'r-', 'lineWidth', 3);
    end

    rectHeight = round(rectHeight);
    rectWidth = round(rectWidth);
    xs = repvec(x : (x+rectWidth-1), rectHeight);
    ys = expandVector(y : (y+rectHeight-1), rectWidth);
    xs = round(xs); % Avoid problems with numerical errors
    ys = round(ys);
    selectedPixels = sub2ind([height width nFrames], ys, xs, repvec(frameId, length(xs)) );

    selectedFeatures = imageFeatures(selectedPixels, :);

    if ~scaledFeatures
      % Cancel out the non-linear scaling
      % which was: tmp = exp((tmp - 1) .* 1000)
      selectedFeatures = (log(selectedFeatures) ./ 1000) + 1;
      selectedFeatures(selectedFeatures(:) == -inf) = 0;
    end
    tmpData = sum(selectedFeatures, 1) ./ length(selectedPixels); % Accumulate histograms over selected pixels

    % Get 2D histograms
    tmpData = reshape(tmpData, nOrientations, nSpeeds);
    %tmpData = tmpData(:, 2 : end); % Remove the "static" row (uniform, not nice)

    %figure; bar3(tmpData);
    %figure; bar3( exp((tmpData - 1) .* 60) );

    % Make the figure
    set(figureIdBars, 'Position', [10 50 300 300]);
    figure(figureIdBars);
    clf;
    bar3(tmpData, 1);
    view(125, 35); % Azimuth, elevation
    ylim(0.5 + [0 nOrientations]);
    xlim(0.5 + [0 nSpeeds]);
    if max(tmpData(:)) > 0
       zlim([0 max(tmpData(:))]);
    end
    adjustFigure();
    %xlabel('Spatial orientation'); ylabel('Velocity');
    drawnow();

    % Display the same data on a polar plot
    if ishandle(figureIdPolarMap) && exist('speeds', 'var') && length(speeds) >= 5
      speed_tmp = speeds;
      speed_tmp(speed_tmp == +inf) = max(speed_tmp(speed_tmp < +inf))+speed_tmp(2)-speed_tmp(1);

      set(figureIdPolarMap, 'Position', [320 50 300 300]);
      figure(figureIdPolarMap);
      clf;
      tmpData(end + 1, :) = tmpData(1, :); % Replicate the first column at the end to "close" (circularly) the plot
      polarplot3d(tmpData', ...
        'PlotType', 'surfn', ...
        'PolarGrid', {1 1}, ...
        'TickSpacing', 0, ...
        'AngularRange',  [ -pi/nOrientations   (2*pi - pi/nOrientations) ], ...
        'RadialRange',  speed_tmp, ...
        'GridScale',  [80 80],  ...
        'InterpMethod', 'nearest', ...
        'AxisLocation', 'off', ...
        'PolarDirection', 'cw');
      view(90, 90); % Azimuth, elevation
      adjustFigure();
      ylim(xlim);
    end

  end

  %------------------------------------------------------------------------
  % Get a new action from the user
  %------------------------------------------------------------------------
  pause(); % Wait for key press
  if ~ishandle(figureId)
    return; % The user closed the main figure
  end
  key = get(gcf, 'CurrentCharacter');
  if isempty(key), continue; end
  switch key
    %----------------------------------------------------------------------
    % Change frame or level
    %----------------------------------------------------------------------
    case 28, frameId = max(frameId - 1, 1); % Left arrow
    case 29, frameId = min(frameId + 1, nFrames); % Right arrow

    %----------------------------------------------------------------------
    case 'r' % Display features of a region
    %----------------------------------------------------------------------
    % Let the user pick a region
    printf('Pick a pixel or draw a rectangular region...\n');
    figure(figureId);
    rect = getrect(figureId);
    x = rect(1);
    y = rect(2);
    rectWidth = rect(3);
    rectHeight = rect(4);
    if rectWidth  < 1,  rectWidth  = 1; end
    if rectHeight < 1,  rectHeight = 1; end
    if ~ishandle(figureIdBars)
      figureIdBars = figure(); % Create an empty figure
    end
    if ~ishandle(figureIdPolarMap)
      figureIdPolarMap = figure(); % Create an empty figure
    end

    %----------------------------------------------------------------------
    case 'v' % Visualize orientation of peak feature
    %----------------------------------------------------------------------
    visualizePeakOrientation = ~visualizePeakOrientation;

    %----------------------------------------------------------------------
    case 'z' % Toggle scaled/unscaled(=raw) feature histograms
    %----------------------------------------------------------------------
    scaledFeatures = ~scaledFeatures;

    %----------------------------------------------------------------------
    case 's' % Save the histogram figure
    %----------------------------------------------------------------------
    if ishandle(figureIdBars)
      figure(figureId); export_fig(sprintf('savedFigures/frame%d.png', frameId));
      figure(figureIdBars); export_fig(sprintf('savedFigures/hist%d.png' , frameId));
      if ishandle(figureIdPolarMap)
        figure(figureIdPolarMap); export_fig(sprintf('savedFigures/polar%d.png', frameId));
      end
    end

    %----------------------------------------------------------------------
    case 'q' % Quit
    %----------------------------------------------------------------------
    close(figureId);
    return;

    otherwise
    continue;
  end % switch key

end % while true

%==========================================================================
function adjustFigure()
%ADJUSTFIGURE Tweak properties of current figure to make plot look better.

set(gcf, 'Color', [1 1 1]); % White background color

axis square;
camproj('orthographic')
box off;
grid off;

% Change colormap
tmp2 = bone(50);
colormap(tmp2(13:end-13, :));

set(gca, 'XTick', []);
set(gca, 'YTick', []);
set(gca, 'ZTick', []);
set(gca, 'XGrid', 'off');
set(gca, 'YGrid', 'off');
set(gca, 'ZGrid', 'off');

drawnow();
