function boundaryMap = makeBoundaryMap(resultPath, thinBoundaries, frameId)
%MAKEBOUNDARYMAP Make boundary map by aggregating results of segmentation of all levels, for one specific frame of the video.
%   BOUNDARYMAP = MAKEBOUNDARYMAP(RESULTPATH, THINBOUNDARIES, FRAMEID) loads
%   the results of the segmentation algorithm for one specific frame
%   (specified, or the middle frame of the video by default), and makes a
%   visualization of the boundaries of the segments appearing in all levels
%   of the segmentation. The strength of a boundary is determined from the
%   highest level this boundary appears in.
%
%   If no output argument is specified, the result is displayed, otherwise
%   it is returned as "probabilistic edge map", i.e. an image of values in
%   [0, 1] indicating the strength of boundaries.
%
%   The result is also saved in two image files: 'Boundaries.png' and
%   'BoundariesComposite.png'.

% Author: Damien Teney

% Declare constants
% Column indices in 'edges'
A = 1; B = 2; WEIGHT = 3;
% Column indices in 'nodes'
MST_WEIGHT = 1; SIZE = 2; RANK = 3; VARIANCE = 4; COLOR = 5; MIN_FRAME = 8; MAX_FRAME = 9; MIN_X = 10; MAX_X = 11; MIN_Y = 12; MAX_Y = 13; HISTOGRAMS = 14;

allBoundaries = {}; % Cell array of images of the boundaries of each level

printf('Loading level ');
for level = 3:100 % For each level (skip the first 2, they always have too many segments)
  printf('%u ', level);
  %------------------------------------------------------------------------
  % Load the segmentation at the current level
  %------------------------------------------------------------------------
  resultFileName = sprintf('%s%sSegmentation_%02u.mat', resultPath, filesep, level);
  if ~exist(resultFileName, 'file')
    if level == 3
      error('Cannot load segmentation !');
    end
    break;
  end
  load(resultFileName, 'pixelAssignment', 'edges');

  %------------------------------------------------------------------------
  % Identify the boundaries at this level
  %------------------------------------------------------------------------
  % Only do this on the middle frame of the sequence
  [height width nFrames] = size(pixelAssignment);
  if nargin < 4
    frameId = 1 + ceil(nFrames / 2); % Pick a frame: middle frame of the sequence
  end

  pixelAssignmentFrame = pixelAssignment(:, :, frameId);
  frameNodes = unique(pixelAssignmentFrame); % Nodes present in the current frame
  if numel(frameNodes) > 100 % Too many segments (will be expensive to compute)
    continue; % Go on with higher levels
  end
  boundaries = cell(0); % Empty initialization
  % Find edges that have at least one node in the current frame
  connectedEdges = false(size(edges, 1), 1); % Binary vector marking these edges
  for a = frameNodes(:)' % For each node of the current frame
    connectedEdges = connectedEdges | edges(:, A) == a | edges(:, B) == a; % Edges connected to the node
  end
  connectedEdgesIds = find(connectedEdges);
  for e = connectedEdgesIds(:)' % For each connected edge
    % Make a mask of the pixels along the edge
    nodeAMask = (pixelAssignmentFrame == edges(e, A)); % Mask of the connected nodes in the current frame
    nodeBMask = (pixelAssignmentFrame == edges(e, B));
    if ~any(nodeAMask(:)) || ~any(nodeBMask(:)) % One of the nodes does NOT appear in the frame
      continue;
    end

    % Get a binary mask of the boundary
    if thinBoundaries
      % Method 1: gives 1 pixel-wide boundaries
      bA = getBoundariesFromSegmentMap(nodeAMask);
      bB = getBoundariesFromSegmentMap(nodeBMask);
      boundaryMask = (bA & nodeBMask) | (bB & nodeAMask);
      boundaryMask = bwmorph(boundaryMask, 'thin', inf); % Thin it even more (4-connectivity instead of 8-connectivity)
    else
      % Method 2: gives 2 pixel-wide boundaries
      nodeAMask = imdilate(nodeAMask, [0 1 0; 1 1 1 ; 0 1 0]);
      nodeBMask = imdilate(nodeBMask, [0 1 0; 1 1 1 ; 0 1 0]);
      boundaryMask = nodeAMask & nodeBMask;
    end

    if ~any(boundaryMask(:)) % The nodes do not connect in the current frame
      continue;
    end

    % Save the result
    boundaries{end + 1}    = struct();
    boundaries{end}.weight = edges(e, WEIGHT);
    boundaries{end}.mask   = boundaryMask;

    %figure(); imshow(nodeAMask); figure(); imshow(nodeBMask); figure(); imshow(boundaryMaskk); % Debug dsplay
  end % For each connected edge

  allBoundaries{end + 1} = zeros(height, width); % Add a new entry
  for b = 1 : length(boundaries) % For each boundary of the current level
    mask = boundaries{b}.mask * boundaries{b}.weight;
    allBoundaries{end} = max(allBoundaries{end}, mask);
  end
  %allBoundaries{end} = allBoundaries{end} .* bwmorph(allBoundaries{end}, 'thin', inf); % Thin it even more (4-connectivity instead of 8-connectivity)
  if max(allBoundaries{end}(:)) ~= 0
    allBoundaries{end} = allBoundaries{end} ./ max(allBoundaries{end}(:)); % Make them values in [0, 1]
  end

end % For each level
printf('\n');

%--------------------------------------------------------------------------
% Make a composite image of boundaries of all levels
%--------------------------------------------------------------------------
resultComposite = assembleImages(fliplr(allBoundaries)); % Assemble, in reverse order (high levels first)

%--------------------------------------------------------------------------
% Get an image with all boundaries of all levels
%--------------------------------------------------------------------------
% Turn 'allBoundaries' into a 3D matrix
allBoundariesMatrix = zeros(length(allBoundaries), height, width);
for level = 1:length(allBoundaries)
  allBoundariesMatrix(level, :, :) = allBoundaries{level};
end

% Aggregate weights of all levels
% First: make the strength of boundaries proportional to the max level they appear in
allBoundariesMatrix2 = allBoundariesMatrix;
for level = 1:length(allBoundaries)
  allBoundariesMatrix2(level, :, :) = allBoundariesMatrix(level, :, :) .* level;
end
boundaryMap = max(allBoundariesMatrix2, [], 1);

% Second: refine the boundary strength by the accumulated weights of each bit of boundary at *all* levels
%%{
boundariesStrengthSum = sum(allBoundariesMatrix, 1);
% Make it values in [0,1[, so that it does not overrides values dependent on the level
boundariesStrengthSum = boundariesStrengthSum ./ (max(boundariesStrengthSum(:)) + eps);
boundaryMap = boundaryMap + boundariesStrengthSum;
%}

boundaryMap = squeeze(boundaryMap); % Remove first (singleton) dimension
boundaryMap = boundaryMap ./ max(boundaryMap(:)); % Scale final values into [0,1]

%--------------------------------------------------------------------------
% Save/display the result
%--------------------------------------------------------------------------
imwrite(boundaryMap.^2, [resultPath filesep 'Boundaries.png']);
imwrite(resultComposite.^2, [resultPath filesep 'BoundariesComposite.png']);

if nargout == 0
  %figure(); imshow(resultComposite, []);
  %figure(); imshow(resultComposite.^2, []);

  %figure(); imshow(boundaryMap, [0 1]);
  figure(); imshow(boundaryMap.^2, [0 1]);
  %figure(); imshow(boundaryMap.^4, [0 1]);

  drawnow();
end

%==========================================================================
function b = getBoundariesFromSegmentMap(segmentMap)
%GETBOUNDARIESFROMSEGMENTMAP From a segmentation map (of a specific level), compute a binary boundary map with 1 pixel wide boundaries.

% Get shifted versions of the segment map
e = zeros(size(segmentMap));
s = zeros(size(segmentMap));
se = zeros(size(segmentMap));
e(:, 1:end-1) = segmentMap(:, 2:end);
s(1:end-1, :) = segmentMap(2:end, :);
se(1:end-1, 1:end-1) = segmentMap(2:end, 2:end);

% Get boundary map
b = (segmentMap ~= e | segmentMap ~= s | segmentMap ~= se);
b(end, :) = (segmentMap(end, :) ~= e(end, :));
b(:, end) = (segmentMap(:, end) ~= s(:, end));
b(end, end) = 0;
