//=========================================================================
// computeEdgeWeights(edges, nodes, HIST_COLUMNS)
// Author: Damien Teney
//=========================================================================

#include <mex.h>
#include "mergeNodes.h"

// Declare constants (minus 1 each time to get 0-based indices, instead of Matlab's 1-based indices)
// Column indices in 'edges'
#define A             (1-1)
#define B             (2-1)
#define WEIGHT        (3-1)

//=========================================================================
// Input arguments:
//   egges: nEdges x 3 matrix, type double
//   nodes: nNodes x M matrix, type double
//   HIST_COLUMNS: nHistograms x 2 matrix, type double
// Output arguments:
//   none, 'edges' is modified in-place
//=========================================================================
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
  // Input
  double *edges, *nodes, *histColumns;
  int nEdges, nNodes, nHistograms;
  // Misc
  int e, h, k; // Loop indices
  int a, b; // Node IDs
  int histColumnStart, histColumnEnd;
  int hLength;
  double h1Sum, h2Sum;
  double h1[400], h2[400]; // Copy of histograms to compare, normalized, hopefully big enough
  double d, dTmp;

  // Check the number of arguments
  if (nrhs < 3)
    mexErrMsgTxt("Wrong number of input arguments !");

  // Get input
  nEdges      = (int) mxGetM(prhs[0]);
  nNodes      = (int) mxGetM(prhs[1]);
  nHistograms = (int) mxGetM(prhs[2]);
  if ((int) mxGetN(prhs[0]) < 3)
    mexErrMsgTxt ("EDGES must be of dimensions Nx3 !");
  if ((int) mxGetN(prhs[2]) != 2)
    mexErrMsgTxt ("HIST_COLUMNS must be of dimensions Nx2 !");
  if (mxGetClassID(prhs[0]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[1]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[2]) != mxDOUBLE_CLASS)
    mexErrMsgTxt ("Invalid data types !");
  edges = (double *) mxGetPr(prhs[0]);
  nodes = (double *) mxGetPr(prhs[1]);
  histColumns = (double *) mxGetPr(prhs[2]);

  //-----------------------------------------------------------------------
  // Main code
  //-----------------------------------------------------------------------
  for (e = 0; e < nEdges; e++) { // For each edge
    a = (int) edges[nEdges * A + e] - 1; // Get 0-based indices
    b = (int) edges[nEdges * B + e] - 1;

    d = 1.0; // Initialize the distance between the two nodes/segments

    for (h = 0; h < nHistograms; h++) { // For each histograms
      histColumnStart = (int) histColumns[nHistograms * 0 + h] - 1; // Get 0-based indices
      histColumnEnd   = (int) histColumns[nHistograms * 1 + h] - 1;
      hLength = histColumnEnd + 1 - histColumnStart;
      if (hLength > 400)
        mexErrMsgTxt("Histograms too big !");

      // Make a copy of the histograms and get the sum of the entries in each histogram
      h1Sum = 0;
      h2Sum = 0;
      for (k = 0; k < hLength; k++) {
        h1[k] = nodes[(histColumnStart + k) * nNodes + a];
        h2[k] = nodes[(histColumnStart + k) * nNodes + b];
        h1Sum += h1[k];
        h2Sum += h2[k];
      }

      // Normalize each histogram, so that each sums to 1 (necessary to compute distance)
      for (k = 0; k < hLength; k++) {
        h1[k] /= h1Sum;
        h2[k] /= h2Sum;
      }

      // Compute the Chi-squared distance between histograms
      // d(h1, h2) = sum( (h1_i - h2_i)^2 / (h1_i + h2_i) ) / 2;
      // Note that it works with concatenated histograms, i.e. the sum of distances between several histograms = the distance between concatenated histograms
      dTmp = 0;
      for (k = 0; k < hLength; k++)
        if ((h1[k] + h2[k]) > 0)
          dTmp += ((h1[k] - h2[k]) * (h1[k] - h2[k])) / (h1[k] + h2[k]);
      dTmp = dTmp / 2.0;

      // Combine several histogram distances
      d = d * (1 - dTmp); // Close to 0 only if they all are; close to 1 if any of them is close to 1
    } // For each histogram

    d = (1 - d);

    // Save the result
    edges[nEdges * WEIGHT + e] = d;
  } // For each edge
}
