function edges = createEdges(inputVideo, connectivity)
%CREATEEDGES Create list of edges between adjacent nodes; 6-connectivity in space/time. Edge weights initialized as pixel color differences).

%   Author: Damien Teney

if connectivity ~= 6 && connectivity ~= 10 && connectivity ~= 26
  error('Invalid connectivity for creating edges !');
end

% Declare constants
% Column indices in 'edges'
A = 1; B = 2; WEIGHT = 3;

height = size(inputVideo, 1);
width = size(inputVideo, 2);
nFrames = size(inputVideo, 3);
nNodes = height * width * nFrames;
nodeIds = reshape(1:nNodes, [height width nFrames]);

% Convert RGB values to Lab values
for frameId = 1:nFrames
  inputVideo(:, :, frameId, :) = applycform(squeeze(inputVideo(:, :, frameId, :)), makecform('srgb2lab'));
  inputVideo(:, :, frameId, 2) = 50 + inputVideo(:, :, frameId, 2) ./ 2; % Get A values in [0,100]
  inputVideo(:, :, frameId, 3) = 50 + inputVideo(:, :, frameId, 3) ./ 2; % Get B values in [0,100]
end
inputVideo = inputVideo ./ 100; % Get all pixel values in (approximately) [0,1]
edges = zeros(0, 3);

%--------------------------------------------------------------------------
% Create edges between pixels: same frame
%--------------------------------------------------------------------------
% Same frame: left neighbour
shift = [0 1 0];
inputVideo_shifted = circshift(inputVideo, [shift 0]);
nodeIds_shifted    = circshift(nodeIds   , shift);
weights = getEdgeWeights(inputVideo, inputVideo_shifted);
nodeIds1 = nodeIds(        1:height, 2:width, 1:nFrames);
nodeIds2 = nodeIds_shifted(1:height, 2:width, 1:nFrames);
weights  = weights(        1:height, 2:width, 1:nFrames);
edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows

% Same frame: upper neighbour
shift = [1 0 0];
inputVideo_shifted = circshift(inputVideo, [shift 0]);
nodeIds_shifted    = circshift(nodeIds   , shift);
weights = getEdgeWeights(inputVideo, inputVideo_shifted);
nodeIds1 = nodeIds(        2:height, 1:width, 1:nFrames);
nodeIds2 = nodeIds_shifted(2:height, 1:width, 1:nFrames);
weights  = weights(        2:height, 1:width, 1:nFrames);
edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows

if connectivity >= 10
  % Same frame: upper-left neighbour
  shift = [1 1 0];
  inputVideo_shifted = circshift(inputVideo, [shift 0]);
  nodeIds_shifted    = circshift(nodeIds   , shift);
  weights = getEdgeWeights(inputVideo, inputVideo_shifted);
  nodeIds1 = nodeIds(        2:height, 2:width, 1:nFrames);
  nodeIds2 = nodeIds_shifted(2:height, 2:width, 1:nFrames);
  weights  = weights(        2:height, 2:width, 1:nFrames);
  edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows

  % Same frame: upper-right neighbour
  shift = [1 -1 0];
  inputVideo_shifted = circshift(inputVideo, [shift 0]);
  nodeIds_shifted    = circshift(nodeIds   , shift);
  weights = getEdgeWeights(inputVideo, inputVideo_shifted);
  nodeIds1 = nodeIds(        2:height, 1:width-1, 1:nFrames);
  nodeIds2 = nodeIds_shifted(2:height, 1:width-1, 1:nFrames);
  weights  = weights(        2:height, 1:width-1, 1:nFrames);
  edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows
end

%--------------------------------------------------------------------------
% Create edges between pixels: previous frame
%--------------------------------------------------------------------------
% Previous frame: same positon
shift = [0 0 1];
inputVideo_shifted = circshift(inputVideo, [shift 0]);
nodeIds_shifted    = circshift(nodeIds   , shift);
weights = getEdgeWeights(inputVideo, inputVideo_shifted);
nodeIds1 = nodeIds(        1:height, 1:width, 2:nFrames);
nodeIds2 = nodeIds_shifted(1:height, 1:width, 2:nFrames);
weights  = weights(        1:height, 1:width, 2:nFrames);
edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows
edges = edges((edges(:, A) > 0 & edges(:, B) > 0), :); % Remove edges built between pixels that were mapped (with the flow) from outside the input images

if connectivity >= 26
  % Previous frame: left neighbour
  shift = [0 1 1];
  inputVideo_shifted = circshift(inputVideo, [shift 0]);
  nodeIds_shifted    = circshift(nodeIds, shift);
  weights = getEdgeWeights(inputVideo, inputVideo_shifted);
  nodeIds1 = nodeIds(        1:height, 2:width, 2:nFrames);
  nodeIds2 = nodeIds_shifted(1:height, 2:width, 2:nFrames);
  weights  = weights(        1:height, 2:width, 2:nFrames);
  edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows

  % Previous frame: upper neighbour
  shift = [1 0 1];
  inputVideo_shifted = circshift(inputVideo, [shift 0]);
  nodeIds_shifted    = circshift(nodeIds, shift);
  weights = getEdgeWeights(inputVideo, inputVideo_shifted);
  nodeIds1 = nodeIds(        2:height, 1:width, 2:nFrames);
  nodeIds2 = nodeIds_shifted(2:height, 1:width, 2:nFrames);
  weights  = weights(        2:height, 1:width, 2:nFrames);
  edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows

  % Previous frame: upper-left neighbour
  shift = [1 1 1];
  inputVideo_shifted = circshift(inputVideo, [shift 0]);
  nodeIds_shifted    = circshift(nodeIds, shift);
  weights = getEdgeWeights(inputVideo, inputVideo_shifted);
  nodeIds1 = nodeIds(        2:height, 2:width, 2:nFrames);
  nodeIds2 = nodeIds_shifted(2:height, 2:width, 2:nFrames);
  weights  = weights(        2:height, 2:width, 2:nFrames);
  edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows

  % Previous frame: upper-right neighbour
  shift = [1 -1 1];
  inputVideo_shifted = circshift(inputVideo, [shift 0]);
  nodeIds_shifted    = circshift(nodeIds, shift);
  weights = getEdgeWeights(inputVideo, inputVideo_shifted);
  nodeIds1 = nodeIds(        2:height, 1:width-1, 2:nFrames);
  nodeIds2 = nodeIds_shifted(2:height, 1:width-1, 2:nFrames);
  weights  = weights(        2:height, 1:width-1, 2:nFrames);
  edges = [ edges ; nodeIds1(:) nodeIds2(:) weights(:) ]; % Append rows
end

%==========================================================================
function weights = getEdgeWeights(inputVideo, inputVideo_shifted)
%GETWEDGEWEIGHTS Compute weights of edges.

% Take the normalized color difference between the original video and the shifted one
weights = sqrt(sum((inputVideo - inputVideo_shifted) .^ 2, 4) / 3); % Normalized color differences accross the 3 channels (4th dimension), scaled to [0, 1]
