//=========================================================================
// mergeNodes_similar(edges, nodes, nodesParents, tau, minNNodes, notBetweenFrames)
// Author: Damien Teney
//=========================================================================

#include <mex.h>
#include "mergeNodes.h"

#define WEIGHT_THRESHOLD 0.0

//=========================================================================
// Input arguments:
//   edges: nEdges x 3 matrix, type double
//   nodes: nNodes x M matrix, type double
//   nodesParents: nNodes vector, type double
//   tau: scalar, type double
//   minNbNodes: scalar, type double
//   notBetweenFrames: boolean
// Output arguments:
//   none, 'edges', 'nodes', and 'nodesParents' are modified in-place
//=========================================================================
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
  // Input
  double *edges, *nodes, *nodesParents;
  int nEdges, nNodes, nNodesCols; // Dimensions of the matrices 'edges' and 'nodes'
  int nNodesActual; // Actual number of nodes, since we do not actually remove rows of the matrix 'nodes', nNodes is constant
  double *tmpPtr;
  double tau;
  double minNNodes;
  mxLogical *tmpPtr2;
  bool notBetweenFrames;
  // Misc
  int e, a, b; // Edge and node IDs
  double edgeWeight;
  double nodeASize, nodeBSize;
  double nodeADuration, nodeBDuration;
  double nodeAMeanArea, nodeBMeanArea;
  double nodeAMaxVolume, nodeBMaxVolume;

  // Check the number of arguments
  if (nrhs < 4)
    mexErrMsgTxt ("Wrong number of input arguments !");

  // Get input
  nEdges      = (int) mxGetM(prhs[0]);
  nNodes      = (int) mxGetM(prhs[1]);
  nNodesCols  = (int) mxGetN(prhs[1]);
  nNodesActual = nNodes;
  if ((int) mxGetN(prhs[0]) < 3)
    mexErrMsgTxt ("EDGES must be of dimensions Nx3 !");
  if ((int) mxGetN(prhs[1]) < HISTOGRAMS)
    mexErrMsgTxt ("NODES does not have enough columns !");
  if ((int) mxGetM(prhs[2]) * mxGetN(prhs[2]) != nNodes)
    mexErrMsgTxt ("NODESPARENTS does not have the same number of elements as NODES !");
  if (mxGetClassID(prhs[0]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[1]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[2]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[3]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[4]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[5]) != mxLOGICAL_CLASS)
    mexErrMsgTxt ("Invalid data types !");
  edges = (double *) mxGetPr(prhs[0]);
  nodes = (double *) mxGetPr(prhs[1]);
  nodesParents = (double *) mxGetPr(prhs[2]);
  tmpPtr = (double *) mxGetPr(prhs[3]); tau = *tmpPtr;
  tmpPtr = (double *) mxGetPr(prhs[4]); minNNodes = *tmpPtr;
  tmpPtr2 = mxGetLogicals(prhs[5]); notBetweenFrames = *tmpPtr2;

  //-----------------------------------------------------------------------
  // Main code
  //-----------------------------------------------------------------------
  for (e = 0; e < nEdges; e++) { // For each edge
    if (nNodesActual <= minNNodes)
      break;

    a = (int) edges[nEdges * A + e] - 1; // Get 0-based indices
    b = (int) edges[nEdges * B + e] - 1;

    // Find the root node of 'a' and 'b'
    a = findRootNode(a, nodesParents);
    b = findRootNode(b, nodesParents);

    // Update edge info with root nodes
    edges[nEdges * A + e] = a + 1; // Restore 1-based indices
    edges[nEdges * B + e] = b + 1;

    if (a == b)
      continue;

    if (notBetweenFrames && (    nodes[MAX_FRAME * nNodes + a] < nodes[MIN_FRAME * nNodes + b]     // Node 'a' ends before node 'b' appears
                              || nodes[MAX_FRAME * nNodes + b] < nodes[MIN_FRAME * nNodes + a] ) ) // Node 'b' ends before node 'a' appears
      continue;

    // Get dimensions of the segments (in number of pixels)
    nodeASize = nodes[SIZE * nNodes + a];
    nodeBSize = nodes[SIZE * nNodes + b];
    nodeADuration = nodes[MAX_FRAME * nNodes + a] - nodes[MIN_FRAME * nNodes + a] + 1;
    nodeBDuration = nodes[MAX_FRAME * nNodes + b] - nodes[MIN_FRAME * nNodes + b] + 1;
    if (nodeASize == 0 || nodeBSize == 0 || nodeADuration == 0 || nodeBDuration == 0) {
      mexPrintf("Edge %u (zero-based index)", e);
      mexErrMsgTxt("Invalid node properties ! (zero size or duration)");
    }
    nodeAMeanArea = nodeASize / nodeADuration; // Average area in each frame (= total number of pixels divided by the the number of frames in which the segment appears)
    nodeBMeanArea = nodeBSize / nodeBDuration;
    nodeAMaxVolume = (nodes[MAX_X * nNodes + a] - nodes[MIN_X * nNodes + a] + 1) * (nodes[MAX_Y * nNodes + a] - nodes[MIN_Y * nNodes + a] + 1) * nodeADuration; // Max number of pixels in the cuboid encompassing the segment
    nodeBMaxVolume = (nodes[MAX_X * nNodes + b] - nodes[MIN_X * nNodes + b] + 1) * (nodes[MAX_Y * nNodes + b] - nodes[MIN_Y * nNodes + b] + 1) * nodeBDuration;
    edgeWeight = edges[WEIGHT * nEdges + e];

    if ( edgeWeight <= WEIGHT_THRESHOLD // Always merge edges of (almost) null weight (even if one of the nodes is marked as ambiguous)
//              || (     edgeWeight <= nodes[MST_WEIGHT * nNodes + a] + tau * exp(-20 * (nodeASize / nodeAMaxVolume))
//                   &&  edgeWeight <= nodes[MST_WEIGHT * nNodes + b] + tau * exp(-20 * (nodeBSize / nodeBMaxVolume)) ) ) {
              || (     edgeWeight <= nodes[MST_WEIGHT * nNodes + a] + tau / nodeAMeanArea
                   &&  edgeWeight <= nodes[MST_WEIGHT * nNodes + b] + tau / nodeBMeanArea ) ) {
      joinNodes(a, b, e, nodes, nodesParents, nNodes, nNodesCols, edges, nEdges);
      nNodesActual--;
    } // If merge
  } // For each edge
}
