//=========================================================================
// mergeNodes_small(edges, nodes, nodesParents, minArea, minWidth, minHeight, minDuration, minNNodes)
// Author: Damien Teney
//=========================================================================

#include <mex.h>
#include "mergeNodes.h"

//=========================================================================
// Input arguments:
//   edges: nEdges x 3 matrix, type double
//   nodes: nNodes x M matrix, type double
//   nodesParents: nNodes vector, type double
//   minArea: scalar, type double
//   minWidth: scalar, type double
//   minHeight: scalar, type double
//   minNNodes: scalar, type double
// Output arguments:
//   none, 'edges', 'nodes', and 'nodesParents' are modified in-place
//=========================================================================
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
  // Input
  double *edges, *nodes, *nodesParents;
  int nEdges, nNodes, nNodesCols; // Dimensions of the matrices 'edges' and 'nodes'
  int nNodesActual; // Actual number of nodes, since we do not actually remove rows of the matrix 'nodes', nNodes is constant
  double *tmpPtr, minArea, minWidth, minHeight, minDuration, minNNodes;
  mxLogical *tmpPtr2;
  // Misc
  int e, a, b; // Edge and node IDs
  double nodeAMeanArea, nodeBMeanArea;
  double nodeAWidth, nodeBWidth, nodeAHeight, nodeBHeight;
  double nodeADuration, nodeBDuration;
  bool joinFlag;

  // Check the number of arguments
  if (nrhs < 8)
    mexErrMsgTxt ("Wrong number of input arguments !");

  // Get input
  nEdges      = (int) mxGetM(prhs[0]);
  nNodes      = (int) mxGetM(prhs[1]);
  nNodesCols  = (int) mxGetN(prhs[1]);
  nNodesActual = nNodes;
  if ((int) mxGetN(prhs[0]) < 3)
    mexErrMsgTxt ("EDGES must be of dimensions Nx3 !");
  if ((int) mxGetN(prhs[1]) < HISTOGRAMS)
    mexErrMsgTxt ("NODES does not have enough columns !");
  if ((int) mxGetM(prhs[2]) * mxGetN(prhs[2]) != nNodes)
    mexErrMsgTxt ("NODESPARENTS does not have the same number of elements as NODES !");
  if (mxGetClassID(prhs[0]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[1]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[2]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[3]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[4]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[5]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[6]) != mxDOUBLE_CLASS
   || mxGetClassID(prhs[7]) != mxDOUBLE_CLASS)
    mexErrMsgTxt ("Invalid data types !");
  edges = (double *) mxGetPr(prhs[0]);
  nodes = (double *) mxGetPr(prhs[1]);
  nodesParents = (double *) mxGetPr(prhs[2]);
  tmpPtr = (double *) mxGetPr(prhs[3]); minArea   = *tmpPtr;
  tmpPtr = (double *) mxGetPr(prhs[4]); minWidth  = *tmpPtr;
  tmpPtr = (double *) mxGetPr(prhs[5]); minHeight = *tmpPtr;
  tmpPtr = (double *) mxGetPr(prhs[6]); minDuration = *tmpPtr;
  tmpPtr = (double *) mxGetPr(prhs[7]); minNNodes = *tmpPtr;

  if (minArea <= 1 && minWidth <= 1 && minHeight <= 1 && minDuration <= 1)
    return; // Nothing to do

  //-----------------------------------------------------------------------
  // Main code
  //-----------------------------------------------------------------------
  for (e = 0; e < nEdges; e++) { // For each edge
    if (nNodesActual <= minNNodes)
      break;

    a = (int) edges[nEdges * A + e] - 1; // Get 0-based indices
    b = (int) edges[nEdges * B + e] - 1;

    // Find the root node of 'a' and 'b'
    a = findRootNode(a, nodesParents);
    b = findRootNode(b, nodesParents);

    // Update edge info with root nodes
    edges[nEdges * A + e] = a + 1; // Restore 1-based indices
    edges[nEdges * B + e] = b + 1;

    if (a == b)
      continue;

    joinFlag = false; // Default value is none of the tests below is passed

    if (nodes[MAX_FRAME * nNodes + a] < nodes[MIN_FRAME * nNodes + b]     // Node 'a' ends before node 'b' appears
     || nodes[MAX_FRAME * nNodes + b] < nodes[MIN_FRAME * nNodes + a] ) { // Node 'b' ends before node 'a' appears
      //---------------------------------------------------------------------
      // The edge is between different frames
      // Merge segments of small duration
      //---------------------------------------------------------------------
      // Get the average size of the nodes (segments) in each frame they appear (= their total number of pixels divided by the the number of frames in which they appear)
      nodeADuration = 1 + nodes[MAX_FRAME * nNodes + a] - nodes[MIN_FRAME * nNodes + a];
      nodeBDuration = 1 + nodes[MAX_FRAME * nNodes + b] - nodes[MIN_FRAME * nNodes + b];

      if (nodeADuration < minDuration || nodeBDuration < minDuration)
        joinFlag = true;

    } else {
      //---------------------------------------------------------------------
      // Edge is inside a same frame
      // Merge segments of small area/size
      //---------------------------------------------------------------------
      // Get the dimensions of the nodes (segments)
      // Average area in each frame they appear = their total number of pixels divided by the the number of frames in which they appear
      nodeAMeanArea = nodes[SIZE * nNodes + a] / (1 + nodes[MAX_FRAME * nNodes + a] - nodes[MIN_FRAME * nNodes + a]);
      nodeBMeanArea = nodes[SIZE * nNodes + b] / (1 + nodes[MAX_FRAME * nNodes + b] - nodes[MIN_FRAME * nNodes + b]);
      // Width and height
      nodeAWidth  = 1 + nodes[MAX_X * nNodes + a] - nodes[MIN_X * nNodes + a];
      nodeBWidth  = 1 + nodes[MAX_X * nNodes + b] - nodes[MIN_X * nNodes + b];
      nodeAHeight = 1 + nodes[MAX_Y * nNodes + a] - nodes[MIN_Y * nNodes + a];
      nodeBHeight = 1 + nodes[MAX_Y * nNodes + b] - nodes[MIN_Y * nNodes + b];

      if (nodeAMeanArea < minArea || nodeBMeanArea < minArea)
        joinFlag = true;
      if (nodeAWidth < minWidth || nodeBWidth  < minWidth)
        joinFlag = true;
      if (nodeAHeight < minHeight || nodeBHeight < minHeight)
        joinFlag = true;
    }

    // Actually join the nodes in any of of the conditions checked above was passed
    if (joinFlag) {
      joinNodes(a, b, e, nodes, nodesParents, nNodes, nNodesCols, edges, nEdges);
      nNodesActual--;
    }
  } // For each edge
}
