function nLevels = segmentVideo(inputFileName, resultPath, p)
%SEGMENTVIDEO Hierarchical video segmentation.
%   NLEVELS = SEGMENTVIDEO(INPUTFILENAME, RESULTPATH, P)
%     INPUTFILENAME        Input video file
%     RESULTPATH           Output directory where to save segmentation results
%
%     P                    Parameters, stored in a structure with the following fields:
%     .mergingPercentiles  Vector of 2 values, in percent, indicating the number of edges likely to be joined at, respectively, the first and upper levels.
%                          These percentages correspond to percentiles of the distribution of edge weights.
%                          A typical value is: [20 1].
%     .minSegmentArea      Enforced minimum area of each supervoxel (on average, in each frame it appears; in pixels / frame)
%     .nHierarchyLevels    Maximum number of hierarchy levels; we may not reach it if all segments get merged before.
%     .initialColorBasedSegmentation  Use pixel color differences as edge weights for the first segmentation level.
%     .featureTypes        Vector of types of features to use.
%     .featuresToUse       Cell array of size P.nHierarchyLevels (if not long enough, the last cell is duplicated).
%                          Each cell correspond to a hierarchy level, and stores a vector of indices corresponding to elements of P.featureTypes to use at that level.
%                          For example:
%                          P.featureTypes = [0 30] and P.featuresToUse = {[1], [1 2]}
%                          will use color only at the first level, then color and optical flow for all other levels.
%     .maxNFrames          Maximum number of frames of the video to process (useful to run faster tests only on the beginning of a sequence); set to +inf to process the whole sequence
%     .connectivity        Connectivity in the graph; number of neighbours connected to a pixel in the video volume (6, 10, or 26)
%     .minNSegments        Minimum number of segments over the whole video; stop the segmentation when it is passed. Set to 0 if no limit is desired.
%     .minSegmentDuration  Minimum number of frames during each segment much appear; set to +inf for the clip duration, or 0 if no restriction is desired.
%     .saveVideos          Save video visualizations of the segmentation, in addition to MAT files.
%
%    Return the ID of the last level reached (starting to count at 1).
%    The results of the segmentation are saved in Mat files, that can be opened with VISUALIZESEGMENTATION().

% Author: Damien Teney

% Check input parameters
if length(p.mergingPercentiles) ~= 2 ...
|| p.minSegmentArea < 0 ...
|| p.nHierarchyLevels < 1 ...
|| ~iscell(p.featuresToUse)
  error('Invalid parameters');
end

warning('off', 'MATLAB:load:variableNotFound');

if length(p.featuresToUse) < p.nHierarchyLevels % Not enough elements in 'p.featuresToUse'
  for i = (length(p.featuresToUse) + 1) : p.nHierarchyLevels
    p.featuresToUse{i} = p.featuresToUse{end}; % Duplicate the last row
  end
end

% Declare constants
% Column indices in 'edges'
A = 1; B = 2; WEIGHT = 3;
% Column indices in 'nodes'
MST_WEIGHT = 1; SIZE = 2; RANK = 3; VARIANCE = 4; COLOR = 5; MIN_FRAME = 8; MAX_FRAME = 9; MIN_X = 10; MAX_X = 11; MIN_Y = 12; MAX_Y = 13; HISTOGRAMS = 14;
HIST_COLUMNS = []; % Will be a matrix of size (nHistograms x 2); the two columns store the first/last bins of each histogram

[inputPath inputFileName inputFileExt] = fileparts(inputFileName);
if isempty(inputPath)
  inputPath = ['.' filesep];
else
  inputPath = [inputPath filesep];
end
delete([resultPath filesep 'Segmentation_*.*']); % Delete possible results of previous runs (the number of levels may be different this time so maybe we won't overwrite everything)

% Load video dimensions from just any type of feature (they are stored with all of them)
load([inputPath inputFileName '_features' int2str(p.featureTypes(1)) '.mat'], 'height', 'width', 'nFrames');
nFrames = min(p.maxNFrames, nFrames);

%--------------------------------------------------------------------------
% Create graph
%--------------------------------------------------------------------------
printf('Creating initial graph...'); tic();
[nodes, nodesParents, pixelAssignment] = createNodes(height, width, nFrames, true, true);
if p.initialColorBasedSegmentation
  color = load([inputPath inputFileName '_features0.mat'], 'features');
  color.features = color.features(:, :, 1:nFrames, :); % Limit the number of frames
  edges = createEdges(color.features, p.connectivity);
else
  edges = createEdges2(height, width, nFrames, p.connectivity);
end
clear color;
printf('             %.2fs, %u nodes, %u edges\n', toc(), size(nodes, 1), size(edges, 1));

%--------------------------------------------------------------------------
% Fill histograms
%--------------------------------------------------------------------------
nPixels = height * width * nFrames;

for f = 1 : length(p.featureTypes) % For each type of image feature to use
  featureType = p.featureTypes(f);
  printf('Filling histograms (features % 2u)...', featureType); tic();

  % Load the data
  d = matfile([inputPath inputFileName '_features' int2str(featureType) '.mat']);
  histSize = size(d, 'histograms'); nHistBins = histSize(2);
  if histSize(1) < nPixels
    error(['Features ' int2str(featureType) ' do not have enough entries !']);
  end
  % Copy the histograms
  HIST_COLUMNS(f, 1) = size(nodes, 2) + 1; % Update the list of column IDs of the histograms
  %if p.maxNFrames == 1 && featureType ~= 0 % Static segmentation
  if p.maxNFrames == 1 % Static segmentation
    tmp = d.histograms(:, :);
    tmp = reshape(tmp, height, width, [], nHistBins);
    tmp = sum(tmp(:, :, :, :), 3) ./ size(tmp, 3); % Accumulate features over all frames
    tmp = reshape(tmp, [height * width], nHistBins);
    nodes(:, (end + 1) : (end + nHistBins)) = tmp; % Add columns
    printf('   %.2fs, %u-dimension histograms (accumulated over all frames for static segmentation)\n', toc(), nHistBins);
  else
    nodes(:, (end + 1) : (end + nHistBins)) = d.histograms(1:nPixels, :); % Add columns
    printf('   %.2fs, %u-dimension histograms\n', toc(), nHistBins);
  end
  HIST_COLUMNS(f, 2) = size(nodes, 2);

  % Discard pixels at the borders of the video volume if they don't have reliable features
  % We practically do not remove these pixels from our data structures, but just remove the edges linked to any of these pixels
  try borderPixels = d.borderPixels; % Try to load 'borderPixels', if defined in the input file
  catch, borderPixels = [];
  end
  if ~isempty(borderPixels)
    borderPixels = d.borderPixels; % Rename so that we can modify it

    borderFrames = squeeze(sum(sum(~borderPixels, 1), 2)) > 0; % Binary vector of size 'nFrames', true for frames that are not made entirely of border pixels
    if p.maxNFrames < length(borderFrames)
      borderFrames = borderFrames(1 : p.maxNFrames);
    end
    nFramesWithoutBorders = find(borderFrames, 1, 'last') + 1 - find(borderFrames, 1, 'first'); % Number of frames that can actually be segmented

    borderPixels = borderPixels(1:nPixels); % Limit the number of frames
    % Find edges linked to a border pixel, on one side or the other
    edgesToRemove = borderPixels(edges(:, A)) ...
                  | borderPixels(edges(:, B));
    printf('Removing %u edges linked to border pixels\n', nnz(edgesToRemove));
    edges = edges(~edgesToRemove, :);

    % Mark all nodes corresponding to border pixels (BUT the first one) as "to remove"
    nodes(borderPixels, SIZE) = 0;
    nodes(borderPixels(1), SIZE) = 1; % Keep this one

    % Assign all border pixels to a same node
    firstBorderPixel = find(borderPixels, 1, 'first'); % ID of the node we will keep to represent the border 
    pixelAssignment(borderPixels) = pixelAssignment(firstBorderPixel);
    nodesParents(borderPixels) = firstBorderPixel;
    nodes(firstBorderPixel, COLOR : COLOR + 2) = [0 0 0]; % Assign black color

    if all(borderPixels(:)), error('All pixels are in the borders !'); end
  else
    nFramesWithoutBorders = nFrames;
  end
  clear d;
end

%--------------------------------------------------------------------------
% Hierarchical merging of nodes
%--------------------------------------------------------------------------
for level = 1 : p.nHierarchyLevels % For each level of the hierarchy
  printf('\nLevel %u\n', level); % Display progress
  if isempty(p.featuresToUse), error('No features selected !'); end
  featuresToUse = p.featuresToUse{level};
  printf('\tIDs of selected features: '); printf('%u ', p.featureTypes(featuresToUse)); printf('\n');

  %edges = randomlyPermuteRows(edges); % For easier debug (it masks the possible effects of the ordering of the edges)

  if level == 2 && p.minSegmentArea > 0
    %----------------------------------------------------------------------
    % Special level
    % Merge small segments
    %----------------------------------------------------------------------
    minSegmentArea = p.minSegmentArea;
    printf('\tMerging nodes (min area: % 3upx)...', round(minSegmentArea)); tic();
    edges = sortrows(edges, +WEIGHT); % Sort edges by increasing weight
    mergeNodes_small(edges, nodes, nodesParents, minSegmentArea, 6, 6, 1, p.minNSegments);
    printf('   %.2fs\n', toc());

  elseif level == 7 && p.minSegmentDuration > 0
    %----------------------------------------------------------------------
    % Special level
    % Merge transient segments
    %----------------------------------------------------------------------
    printf('\tComputing edge weights...'); tic();
    computeEdgeWeights(edges, nodes, HIST_COLUMNS(featuresToUse, 1:2));
    printf('             %.2fs\n', toc());

    p.minSegmentDuration = min(p.minSegmentDuration, nFramesWithoutBorders);
    printf('\tMerging transient nodes (min duration: % 3u frames)...', round(p.minSegmentDuration)); tic();
    edges = sortrows(edges, +WEIGHT); % Sort edges by increasing weight
    %mergeNodes_transient(edges, nodes, nodesParents, p.minSegmentDuration, p.minNSegments); % TODO DELETE
    mergeNodes_small(edges, nodes, nodesParents, 1, 1, 1, p.minSegmentDuration, p.minNSegments);
    printf('   %.2fs\n', toc());

  else
    %----------------------------------------------------------------------
    % "Normal" levels: compute weights of edges and merge similar nodes
    %----------------------------------------------------------------------
    if level == 1 && p.initialColorBasedSegmentation
      % Use initial edge weights computed from the pixel differences (not histogram distances)
      printf('\tFeatures actually used: pixel colors\n');
      % Nothing to do, edge weights already set by createEdges()

    else
      % Compute edge weights as differences between histograms of selected features
      printf('\tComputing edge weights...'); tic();
      computeEdgeWeights(edges, nodes, HIST_COLUMNS(featuresToUse, 1:2));
      printf('             %.2fs\n', toc());

      if all(edges(:, WEIGHT) == 0) % Can happen in the last few levels when we use motion only, but remaining segments have too many "empty pixels" and the edge weights thus get 0 weights
        break;
      end
    end

    % Automatically determine 'tau'
    edges = sortrows(edges, +WEIGHT); % Sort edges by increasing weight
    aAreas = nodes(edges(:, A), SIZE) ./ (nodes(edges(:, A), MAX_FRAME) - nodes(edges(:, A), MIN_FRAME) + 1);
    bAreas = nodes(edges(:, B), SIZE) ./ (nodes(edges(:, B), MAX_FRAME) - nodes(edges(:, B), MIN_FRAME) + 1);
    if level == 1
      tau = prctile( max(edges(:, WEIGHT) .* aAreas, edges(:, WEIGHT) .* bAreas), p.mergingPercentiles(1) );
    elseif nRemovedNodes == 0 % The previous iteration did not remove (=join) any nodes
      tau = tau * 1.1;
    else
      tau = prctile( max(edges(:, WEIGHT) .* aAreas, edges(:, WEIGHT) .* bAreas), p.mergingPercentiles(2) );
    end

    % Debug display
    %if level > 1, figure(); bar(sort(edges(:, WEIGHT))); pause(); end % Debug display (weights of all edges)
    %figure(); hist(edges(:, WEIGHT),20);
    %figure(); hist( max(edges(:, WEIGHT) .* aAreas, edges(:, WEIGHT) .* bAreas) );
    %percentile = 100 * find( max(edges(:, WEIGHT) .* aAreas, edges(:, WEIGHT) .* bAreas) > tau, 1) / size(edges, 1)

    if size(nodes, 1) > 20 % Lots of nodes left
      nodes(:, MST_WEIGHT) = 0;
      printf('\tMerging nodes (tau: %.3f)...', tau); tic();
      mergeNodes_similar(edges, nodes, nodesParents, tau, p.minNSegments, false);
      printf('        %.2fs\n', toc());

      if level > 1
        % Automatically determine 'minSegmentArea'
        as = (nodes(:, SIZE) > 1); % Binary vector of "valid" nodes (do not include the nodes from the border for example, which never get joined to others
        nodesAreas = nodes(as, SIZE) ./ (1 + nodes(as, MAX_FRAME) - nodes(as, MIN_FRAME));
        %minSegmentArea = mean(nodesAreas);
        minSegmentArea = (0.5 * mean(sqrt(nodesAreas))) ^ 2;
        %minSegmentArea = ceil(minSegmentArea * 1.2); % Update parameter for the next level
        printf('\tMerging nodes (min area: % 3upx)...', round(minSegmentArea)); tic();
        edges = sortrows(edges, +WEIGHT); % Sort edges by increasing weight
        mergeNodes_small(edges, nodes, nodesParents, minSegmentArea, 1, 1, 1, p.minNSegments);
        printf('   %.2fs\n', toc());
      end

    else % Just a few nodes left: merge them one by one
      printf('\tMerging nodes (one at a time)...', tau); tic();
      mergeNodes_fixedNumber(edges, nodes, nodesParents, 1, p.minNSegments);
      printf('      %.2fs\n', toc());
    end
  end

  %------------------------------------------------------------------------
  % Clean up data structures and save the results of the current level
  %------------------------------------------------------------------------
  printf('\tCleaning up data structures...'); tic();
  previousNNodes = size(nodes, 1);
  edges = edges((edges(:, A) ~= edges(:, B)), :); % Remove edges that have been merged (looping on a same node)
  [edges, nodes, nodesParents, pixelAssignment] = removeChainedLinks(edges, nodes, nodesParents, pixelAssignment);
  [edges, nodes, nodesParents, pixelAssignment] = removeUnusedNodes(edges, nodes, nodesParents, pixelAssignment);
  nRemovedNodes = size(nodes, 1) - previousNNodes;
  edges = removeDuplicateEdges(edges, nodesParents);
  printf('        %.2fs, %u nodes, %u edges\n', toc(), size(nodes, 1), size(edges, 1));

  if size(edges, 1) == 0 || size(nodes, 1) == 1 % All segments have been merged into one, don't even save the result
    break;
  end

  printf('\tGenerating output...'); tic();
  if p.saveVideos
    generateOutputVideos(pixelAssignment, nodes, resultPath, level);
  end
  generateOutputMatFiles(pixelAssignment, nodes, edges, resultPath, level, HIST_COLUMNS)
  nLevels = level; % Update the value that will be returned
  printf('                  %.2fs\n', toc());

  if size(nodes, 1) <= p.minNSegments % We reached or passed the desired number of segments
    break;
  end

end % for each level
