function visualizeSegmentation(resultPath, level)
%VISUALIZESEGMENTATION Visualize results of segmentation (Mat file).

% Author: Damien Teney

% Declare constants
% Column indices in 'edges'
A = 1; B = 2; WEIGHT = 3; EDGE_HISTOGRAMS = 4;
% Column indices in 'nodes'
MST_WEIGHT = 1; SIZE = 2; RANK = 3; VARIANCE = 4; COLOR = 5; MIN_FRAME = 8; MAX_FRAME = 9; MIN_X = 10; MAX_X = 11; MIN_Y = 12; MAX_Y = 13; HISTOGRAMS = 14;

% Initialize
if nargin < 2
  level = 1; % Default value
end

nodesColors = zeros(0, 3);
key = [];
x = []; y = [];
resultFileName = [];

figureId = -1;
figureIdHistograms = -1;
figureIdHistogramsEdge = -1;
figureIdWeights = -1;
figureIdVariance = -1;
figureIdBoundaries = -1;

%close all;
figureId = figure(); % Create an empty figure

printf('Press a key:\n');
printf('\tleft/right Previous/next frame\n');
printf('\tup/down  Higher/lower hierarchy level)\n');
printf('\tu/d      Higher/lower hierarchy level (move by 10 levels at a time)\n');
printf('\ts        Details of a selected segment\n');
printf('\te        Details of an edge between two selected segments\n');
printf('\tE        Display all edges (warning: slow if many segments !)\n');
printf('\tw        Weights of all edges\n');
printf('\tb        Map of boundaries in the current frame\n');
printf('\tq        Quit the visualization tool\n');

while true
  %------------------------------------------------------------------------
  % Load the data
  %------------------------------------------------------------------------
  if isempty(key) || key == 30 || key == 31 || key == 'u' || key == 'd' % First iteration, or just changed the level
    % Load the data
    resultFileName = sprintf('%s%sSegmentation_%02u.mat', resultPath, filesep, level);
    if exist(resultFileName, 'file')
      load(resultFileName, 'pixelAssignment', 'nodes', 'edges', 'HIST_COLUMNS');
    else
      printf('Cannot load: %s\n', resultFileName);
      %warning(['Cannot load: ' resultFileName]);
      if isempty(key) % First iteration
        return; % Stop here
      end
      % Otherwise keep going, maybe the user just changed the level and it just does not exist
    end
    [height width nFrames] = size(pixelAssignment);
    nNodes = size(nodes, 1);
    frameId = floor(nFrames / 2); % Start at the middle of the video


    % Get RGB colors of all segments
    nodesColors = nodes(:, COLOR : (COLOR + 2));
  end

  set(figureId, 'Name', ['Level ' int2str(level) ' - frame ' int2str(frameId)], 'NumberTitle', 'off');
  frameId = min(frameId, size(pixelAssignment, 3)); % Avoid errors if we were on a frameId > 1 on a high level, then we go down to e.g. level 0 (ground truth) wich stores less frames
  visualization = ind2rgb(pixelAssignment(:, :, frameId), nodesColors(:, 1:3)); % Turn the "indexed image" (pixelAssignment) into an RGB image
  figure(figureId); imshow(visualization);

  %------------------------------------------------------------------------
  % Display histograms of a segment
  %------------------------------------------------------------------------
  if ishandle(figureIdHistograms) % The user has picked a pixel and histograms are displayed
    figure(figureId); hold on; plot(x, y, 'bo'); % Mark the pixel picked by the user
    nodeId = pixelAssignment(y, x, frameId); % Get the ID of the node to which the pixel belongs

    % Display misc info on the node/segment
    printf('Segment #%d/%d\n', nodeId, nNodes);
    printf('\tFirst-last frames in which it appears: %d-%d\n', nodes(nodeId, MIN_FRAME), nodes(nodeId, MAX_FRAME));
    printf('\tTotal size: %d px\n', nodes(nodeId, SIZE));
    printf('\tMean size (2D area) per frame: %d px\n', round(nodes(nodeId, SIZE) / (1 + nodes(nodeId, MAX_FRAME) - nodes(nodeId, MIN_FRAME))));

    % Load the histograms of the pixel picked by the user
    nHistograms = size(HIST_COLUMNS, 1);
    h = cell(nHistograms, 1);
    for i = 1:nHistograms
      h{i} = nodes(nodeId, HIST_COLUMNS(i, 1) : HIST_COLUMNS(i, 2));
      maxHistogramValue(i) = max(max(nodes(:, HIST_COLUMNS(i, 1) : HIST_COLUMNS(i, 2) )));
    end
    maxHistogramValue(maxHistogramValue == 0) = 1; % Replace '0' values by '1's

    printf('\tSum of histograms (should be == 1): ');
    for i = 1:nHistograms, printf(' %.2f  ', sum(h{i})); end
    printf('\n');
    printf('\tVariance of histograms within the segment: %f\n', nodes(nodeId, VARIANCE));
    printf('\n');

    % Display the histograms
    figure(figureIdHistograms); % Activate the existing figure
    set(figureIdHistograms, 'Name', 'Feature histograms', 'NumberTitle', 'off'); % Update title
    for i = 1:nHistograms
      subplot(nHistograms, 1, i); bar(h{i}, 1); xlim([0.5 length(h{i}) + 0.5]); ylim([0 maxHistogramValue(i)]);
    end
  end

  %------------------------------------------------------------------------
  % Display edge weights
  %------------------------------------------------------------------------
  if ishandle(figureIdWeights) % The figure already exists
    figure(figureIdWeights); % Bring the figure to the front
    set(figureIdWeights, 'Name', 'Edge weights', 'NumberTitle', 'off'); % Update title
    % Display the data
    weights = edges(:, WEIGHT);
    bar(sort(weights));
  end

  figure(figureId); % Bring the main figure to the front
  drawnow();

  %------------------------------------------------------------------------
  % Get a new action from the user
  %------------------------------------------------------------------------
  %printf('Press a key: left/right (previous/next frame), up/down (higher/lower hierarchy level), h (display histograms), e (display edge weight)...\n');
  pause(); % Wait for key press
  if ~ishandle(figureId)
    return; % The user closed the main figure
  end
  key = get(gcf, 'CurrentCharacter');
  if isempty(key), continue; end
  switch key
    %----------------------------------------------------------------------
    % Change frame or level
    %----------------------------------------------------------------------
    case 28, frameId = max(frameId - 1, 1); % Left arrow
    case 29, frameId = min(frameId + 1, nFrames); % Right arrow
    case 30, level = level + 1; % Up arrow
    case 31, level = max(level - 1, 0); % Down arrow
    case 'u', level = level + 10;
    case 'd', level = max(level - 10, 0);

    %----------------------------------------------------------------------
    case 's' % Display segment details
    %----------------------------------------------------------------------
    % Let the user pick a pixel
    printf('Pick a segment...\n');
    figure(figureId);
    [x y] = ginput(1);
    x = max(min(round(x), size(pixelAssignment, 2)), 0); % Round and cap to valid range of values
    y = max(min(round(y), size(pixelAssignment, 1)), 0);
    if ~ishandle(figureIdHistograms)
      figureIdHistograms = figure(); % Create an empty figure
    end

    %----------------------------------------------------------------------
    case 'e' % Display edge details
    %----------------------------------------------------------------------
    % Let the user pick two pixels, then show the weight of the edge between their segments
    printf('Pick two different segments...\n');
    figure(figureId);
    [xs ys] = ginput(2);
    xs = max(min(round(xs), size(pixelAssignment, 2)), 0); % Round and cap to valid range of values
    ys = max(min(round(ys), size(pixelAssignment, 1)), 0);
    nodeId1 = pixelAssignment(ys(1), xs(1), frameId); % Get the ID of the node to which the pixel belongs
    nodeId2 = pixelAssignment(ys(2), xs(2), frameId); % Get the ID of the node to which the pixel belongs
    edgeId = find((edges(:, 1) == nodeId1 & edges(:, 2) == nodeId2) | (edges(:, 1) == nodeId2 & edges(:, 2) == nodeId1), 1);
    if nodeId1 == nodeId2
      printf('\tYou picked two times the same segment !\n');

    elseif isempty(edgeId)
      printf('\tNo edge found between those segments !\n');

    else % Edge found between segments picked by the user
      printf('\tEdge weight: %f\n', edges(edgeId, 3));
      figure(figureId); hold on; plot(xs, ys, 'r.-'); % Mark the pixels picked by the user
      drawnow(); pause(0.5);

      if size(edges, 2) >= EDGE_HISTOGRAMS
        if ~ishandle(figureIdHistogramsEdge)
          figureIdHistogramsEdge = figure(); % Create an empty figure
        else
          figure(figureIdHistogramsEdge); % Activate existing figure
        end
        % Display feature histogram of the edge
        % Load the histograms of the edge picked by the user
        hEdge = edges(edgeId, EDGE_HISTOGRAMS : end);

        % Get the scale of this type of histograms from those of the nodes
        maxHistogramValueEdge = max(max(nodes(:, HIST_COLUMNS(end, 1) : HIST_COLUMNS(end, 2) )));
        maxHistogramValueEdge(maxHistogramValueEdge == 0) = 1; % Replace '0' values by '1's

        printf('\tSum of edge histogram (should be == 1): %.2f\n', sum(hEdge));

        % Display the histogram
        figure(figureIdHistogramsEdge); % Activate the existing figure
        set(figureIdHistogramsEdge, 'Name', 'Feature histograms of edge', 'NumberTitle', 'off'); % Update title
        bar(hEdge, 1); xlim([0.5 length(hEdge) + 0.5]); ylim([0 maxHistogramValueEdge]);
      end % If edge found between segments picked by the user
    end
    printf('\n');

    %----------------------------------------------------------------------
    case 'E' % Display all edges in the current frame
    %----------------------------------------------------------------------
    % Display edges in the image
    figure(figureId); hold on;
    printf('\tDisplaying all edges between segments in the current frame (warning: slow if many segments !)... ')
    pixelAssignment_frame = pixelAssignment(:, :, frameId);
    edges = edges(edges(:, A) ~= edges(:, B), :); % Remove invalid edges
    for e = 1:size(edges, 1) % For each edge
      % Find coordinates of pixels in the current frame connected by this edge
      [ais ajs] = find(pixelAssignment_frame == edges(e, A));
      [bis bjs] = find(pixelAssignment_frame == edges(e, B));
      if ~isempty(ais) && ~isempty(bis)
        aIJ = mean([ais ajs], 1); % Keep the mean (centroid) of the connected segments
        bIJ = mean([bis bjs], 1); % Keep the mean (centroid) of the connected segments
        plot([aIJ(2) bIJ(2)], [aIJ(1) bIJ(1)], 'k.-'); drawnow();
      end
    end
    printf('done\n');
    printf('\tPress any key to go back to the main menu...\n'); pause();

    %----------------------------------------------------------------------
    case 'w' % Display distribution of all edge weights
    %----------------------------------------------------------------------
    % Display all edge weights
    if ishandle(figureIdWeights) % The figure already exists
      figure(figureIdWeights); % Bring the figure to the front
    else
      figureIdWeights = figure(); % Create a new figure
    end

    %----------------------------------------------------------------------
    case 'b' % Display segment boundaries of the current frame/level
    %----------------------------------------------------------------------
    pixelAssignmentFrame = pixelAssignment(:, :, frameId);
    frameNodes = unique(pixelAssignmentFrame); % Nodes present in the current frame      
    boundaries = cell(0); % Empty initialization
    % Find edges that have at least one node in the current frame
    connectedEdges = false(size(edges, 1), 1); % Binary vector marking these edges
    for a = frameNodes(:)' % For each node of the current frame
      connectedEdges = connectedEdges | edges(:, A) == a | edges(:, B) == a; % Edges connected to the node
    end
    connectedEdgesIds = find(connectedEdges);
    for e = connectedEdgesIds(:)' % For each connected edge
      % Make a mask of the pixels along the edge
      nodeAMask = (pixelAssignmentFrame == edges(e, A)); % Mask of the connected nodes in the current frame
      nodeBMask = (pixelAssignmentFrame == edges(e, B));
      if ~any(nodeAMask(:)) || ~any(nodeBMask(:)) % One of the nodes does NOT appear in the frame
        continue;
      end

      % Save the result
      boundaries{end + 1} = struct();
      boundaries{end}.nodeAMask = nodeAMask;
      boundaries{end}.nodeBMask = nodeBMask;
      boundaries{end}.nodeAFeatures = nodes(edges(e, A), HIST_COLUMNS(end, 1) : HIST_COLUMNS(end, 2));
      boundaries{end}.nodeBFeatures = nodes(edges(e, B), HIST_COLUMNS(end, 1) : HIST_COLUMNS(end, 2));
      boundaries{end}.weight = edges(e, WEIGHT);

      nodeAMask = imdilate(nodeAMask, ones(2)); % Gives 1 pixel-wide boundaries
      nodeBMask = imdilate(nodeBMask, ones(2));
      %nodeAMask = imdilate(nodeAMask, [0 1 0; 1 1 1 ; 0 1 0]); % Gives 2 pixel-wide boundaries
      %nodeBMask = imdilate(nodeBMask, [0 1 0; 1 1 1 ; 0 1 0]);
      edgeMask = nodeAMask & nodeBMask;
      boundaries{end}.mask = edgeMask; % Save the result
      if ~any(edgeMask(:)) % The nodes do not connect in the current frame
        continue;
      end
      %figure(); imshow(nodeAMask); figure(); imshow(nodeBMask); figure(); imshow(edgeMask); % Debug dsplay
    end % For each connected edge

    allBoundaries = zeros(height, width);
    for b = 1 : length(boundaries)
      mask = boundaries{b}.mask * boundaries{b}.weight;
      allBoundaries = max(allBoundaries, mask);
      %allBoundaries = allBoundaries + mask; % Much worse than the previous one (because of a few points where several boundaries coincide)
    end

    if ~ishandle(figureIdBoundaries)
      figureIdBoundaries = figure(); % Create new figure
    else
      figure(figureIdBoundaries); % Activate existing figure
    end
    set(figureIdBoundaries, 'Name', ['Boundaries of segments (frame ' int2str(frameId)  ')'], 'NumberTitle', 'off'); % Update title
    imshow(allBoundaries, []);
    %imshow(allBoundaries .^ 2, []);

    %----------------------------------------------------------------------
    case 'q' % Quit
    %----------------------------------------------------------------------
    close(figureId);
    return;

    otherwise
    continue;
  end % switch key

end % while true
